package ru.nsu.sberlab.security;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import ru.nsu.sberlab.dao.UserRepository;
import ru.nsu.sberlab.exception.FailedUserCreationException;
import ru.nsu.sberlab.model.dto.user.JwtTokenDto;
import ru.nsu.sberlab.model.dto.user.LoginDto;
import ru.nsu.sberlab.model.dto.user.RefreshTokenDto;
import ru.nsu.sberlab.model.dto.user.UserRegistrationDto;
import ru.nsu.sberlab.model.entity.User;
import ru.nsu.sberlab.model.mapper.UserMapper;
import ru.nsu.sberlab.model.util.SocialNetworksConverter;

import java.util.Optional;

@Slf4j
@Service
@RequiredArgsConstructor
public class SecurityService {
    private final AuthenticationManager authenticationManager;
    private final SocialNetworksConverter socialNetworksConverter;
    private final UserDetailsService userDetailsService;
    private final TokenService tokenService;
    private final UserRepository userRepository;
    private final UserMapper userMapper;

    public JwtTokenDto authenticate(LoginDto loginDto) {
        UsernamePasswordAuthenticationToken authenticationToken =
                new UsernamePasswordAuthenticationToken(loginDto.getEmail(), loginDto.getPassword());
        Authentication auth = authenticationManager.authenticate(authenticationToken);
        User user = (User) auth.getPrincipal();
        String accessToken = tokenService.generateAccessToken(user);
        String refreshToken = tokenService.generateRefreshToken(user);
        return new JwtTokenDto(accessToken, refreshToken);
    }

    @Transactional
    public JwtTokenDto register(UserRegistrationDto registrationDto) {
        Optional<User> currentUser = userRepository.findByEmail(registrationDto.getEmail());
        if (currentUser.isPresent() && currentUser.get().isActive()) {
            throw new FailedUserCreationException("api.server.error.user-not-created");
        }
        log.info("Received request to create user with payload: {}", registrationDto);
        User user = userMapper.apply(registrationDto);
        User savedUser = userRepository.save(user);
        savedUser.setUserSocialNetworks(
                socialNetworksConverter.convertSocialNetworksDtoToSocialNetworks(
                        registrationDto.getSocialNetworks(),
                        savedUser
                )
        );
        String accessToken = tokenService.generateAccessToken(savedUser);
        String refreshToken = tokenService.generateRefreshToken(savedUser);
        return new JwtTokenDto(accessToken, refreshToken);
    }

    public JwtTokenDto refreshToken(RefreshTokenDto refreshTokenDto) {
        String email = tokenService.parseToken(refreshTokenDto.getRefreshToken());
        User user = (User) userDetailsService.loadUserByUsername(email);
        String accessToken = tokenService.generateAccessToken(user);
        String refreshToken = tokenService.generateRefreshToken(user);
        return new JwtTokenDto(accessToken, refreshToken);
    }
}
